// $Id: PhaseMgr.cc,v 1.1.2.2 1998/10/16 02:20:44 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include "PhaseMgr.h"
#include "Clock.h"
#include "Log.h"
#include "globals.h"

extern RateMeter TheSubmMeter;
extern RateMeter TheComplMeter;
extern RateMeter TheErrMeter;


/* MeterShot */

void MeterShot::reset() {
	theSubmMeter.reset();
	theComplMeter.reset();
	theErrMeter.reset();
	theShotTime = Time();
}

void MeterShot::shoot() {
	theSubmMeter = TheSubmMeter;
	theComplMeter = TheComplMeter;
	theErrMeter = TheErrMeter;
	theShotTime = max(TheComplMeter.lastTime(), (Time)TheClock);
}

MeterShot &MeterShot::operator -=(const MeterShot &s) {
	theSubmMeter -= s.subm();
	theComplMeter -= s.compl();
	theErrMeter -= s.err();
	return *this;
}

MeterShot operator -(const MeterShot &s1, const MeterShot &s2) {
	MeterShot diff(s1);
	diff -= s2;
	return diff;
}

/* PhaseMeter */

bool PhaseMeter::done() const {
	return theGoal > 0 && TheComplMeter.count() >= theGoal;
}

void PhaseMeter::incGoal(int delta) {
	int g = theGoal + delta;
	while (g + delta <= TheComplMeter.count())
		g += delta;
	goal(g);
}

void PhaseMeter::report(ostream &os) const {
	const MeterShot diff = theFinishShot - theStartShot;
	os << "#" << theName 
		<< " phase done in " << (theFinishShot.time()-theStartShot.time())
		<< " after " << theFinishShot.compl().count() 
		<< " compl ( " << diff.compl() << " compl/sec )"
		<< endl;

	os << theName
		<< "\t " << theFinishShot.time().sec()
		<< "\t " << setw(6) << diff.subm().count()
		<< "\t " << setw(6) << diff.subm()
		<< "\t " << setw(6) << diff.compl().count()
		<< "\t " << setw(6) << diff.compl()
		<< "\t " << setw(6) << diff.err().count()
		<< endl;
}


/* PhaseMgr */

PhaseMgr::PhaseMgr(): thePhase(phNone) {
	thePhases[phWarmUp].name("warm-up");
	thePhases[phMeasure].name("measurement");
	thePhases[phCoolOff].name("cool-off");

	thePinger.name("1K");
}

// set goals (expected number of completions)
void PhaseMgr::goal(int aGoal) {
	assert(thePhase == phNone);

	thePhases[phWarmUp].goal(aGoal/10);          // stop at  10%
	thePhases[phMeasure].goal(aGoal - aGoal/10); //      at  90%
	thePhases[phCoolOff].goal(aGoal);            //      at 100%

	thePinger.goal(1000);

	theGoal = aGoal;
}

// checks metering phase and updates the PhaseMgr as needed
void PhaseMgr::checkPhase() {
	if (thePhase == phNone)
		nextPhase();
	else
	if (thePhases[thePhase].done())
		nextPhase();
	
	if (thePinger.done()) {
		thePinger.stop();
		thePinger.report(Report);
		thePinger.incGoal(1000);
		thePinger.start();
	}
}

MeterShot PhaseMgr::mshot() const {
	PhaseMeter phm = thePhases[max(0,min((int)thePhase,(int)phMeasure))];
	phm.stop();
	return phm.diff();
}

void PhaseMgr::nextPhase() {

	// report results for the current phase if any
	if (thePhase > phNone) {
		thePhases[thePhase].stop();
		thePhases[thePhase].report(Report);
	} else {
		thePinger.start();
	}
	// advance to the next phase if possible
	if (thePhase < phCoolOff) {
		thePhase = (Phase) (1 + (int)thePhase);
		thePhases[thePhase].start();
	} else {
		thePhases[thePhase].goal(-1);
	}
}
