/* $Id: Socket.cc,v 1.18.2.1 1998/09/22 07:43:42 rousskov Exp $ */

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Socket.h"

#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "Bug.h"

bool Socket::create() {
	return create(PF_INET, SOCK_STREAM, 0);
}

bool Socket::create(int domain, int type, int protocol) {
	theFD = ::socket(domain, type, protocol);
	return theFD >= 0;
}

bool Socket::connect(const char *host, int port) {

	struct sockaddr_in s;
	memset(&s, 0, sizeof(s));

	s.sin_family = AF_INET;
	s.sin_addr.s_addr = inet_addr(host);
	s.sin_port = htons(port);

	if (::connect(theFD, (struct sockaddr *) &s, sizeof(s)) < 0) {
		if (errno != EINPROGRESS) { // XXX: || isBlocking()) {
			cerr << here << "fd: " << theFD << " connect from port " << this->port() << " to " << host << ':' << port << " failed";
			if (errno != EADDRINUSE)
				cerr << bug;
			else
				cerr << endl << "cause(?): Address already in use; port: " << this->port() << endl;
			return false;
		}
	}

	return true;
}

bool Socket::bind(int port) {
	static struct sockaddr_in S;

	S.sin_family = AF_INET;
	S.sin_port = htons(port);
	S.sin_addr.s_addr = htonl(INADDR_ANY);

	if (::bind(theFD, (struct sockaddr *) &S, sizeof(S)) != 0) {
		cerr << "bind to port " << port << " failed." << endl;
		if (errno != EADDRINUSE)
			cerr << bug;
		else
			cerr << endl << "cause(?): Address already in use" << endl;
		return false;
	}
	return true;
}

bool Socket::listen(int port) {

	if (!bind(port))
		cerr << "bind failed" << bug;

	if (::listen(theFD, FD_SETSIZE) < 0)
		cerr << "listen failed" << bug;

	return true;
}

Socket Socket::accept(struct sockaddr *addr_p, int *addr_len_p) {
	Socket s;
	struct sockaddr addr;
	int addr_len = sizeof(addr);
	if (!addr_p) addr_p = &addr;
	if (!addr_len_p) addr_len_p = &addr_len;
	s.theFD = ::accept(theFD, addr_p, addr_len_p);
	return s;
}


bool Socket::close() {

	const int res = ::close(theFD);
	theFD = -1;
	return res == 0;
}

int Socket::flags() const {

	int flags = fcntl(theFD, F_GETFL);

	if (flags < 0)
		cerr << here << "get socket flags failed" << bug;

	return flags;
}

int Socket::port() const {
	struct sockaddr_in S;
	int len = sizeof(S);

	if (getsockname(theFD, (struct sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

bool Socket::peerAddr(struct sockaddr *S, int &len) const {
	return getpeername(theFD, S, &len) == 0;
}

void Socket::reusePort(bool set) {
#ifdef SO_REUSEPORT
	setOpt(SO_REUSEPORT, set);
#else
	set = false; /* make compiler happy */
	/* XXX: but no warning? */
	#warning SO_REUSEPORT is not supported on this system
#endif
}

void Socket::linger(int delay) {
#ifdef SO_LINGER
	static struct linger l = { 1, delay };
	setOpt(SO_LINGER, &l, sizeof(l));
#else
	// XXX: but no run-time warning?
	#warning SO_LINGER is not supported on this system
#endif
}

void Socket::blocking(bool set) {

	int flags = this -> flags();

	if (set)
		flags &= ~O_NONBLOCK;
	else
		flags |= O_NONBLOCK;

	if (fcntl(theFD, F_SETFL, flags) == -1)
		cerr << here << "set socket flags failed" << bug;
}

bool Socket::isBlocking() const {
	return flags() & O_NONBLOCK == 0;
}


bool Socket::setOpt(int optName, void *optVal, int optLen) {
	// XXX: maybe assert?
	if (::setsockopt(theFD, SOL_SOCKET, optName, optVal, optLen) != 0) {
		cerr << here << "fd " << theFD << " failed to change opt " << optName << " cause(?): " << strerror(errno) << endl;
		return false;
	}
	return true;
}

bool Socket::setOpt(int optName, bool set) {
	int optVal = (int)set;
	return setOpt(optName, &optVal, sizeof(optVal));
}
