// $Id: filler.cc,v 1.6.2.2 1998/09/22 07:43:49 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include "Time.h"
#include "Size.h"
#include "Rate.h"
#include "Array.h"
#include "Rnd.h"
#include "Gadgets.h"

#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h> /* for strerror */
#include <stdio.h> /* for snprintf */

#include <iostream.h>
#include <iomanip.h>


/*
 * Fills disk with empty files
 * Follows Squid allocation pattern (assuming no releases)
 */


// (count, size) union
class Level {
	public:
		Level(): theCount(0), theSize(0) {}

		void inc(Size sz) { theCount++; theSize += sz; }
		void inc(int count, Size sz) { theCount += count; theSize += sz; }

		bool operator <(const Level &l) { return (theCount >= 0 && l.theCount >= 0 && theCount < l.theCount) || (theSize >= 0 && l.theSize >= 0 && theSize < l.theSize) || (l.theSize < 0 && l.theCount < 0); }

	public:
		int theCount;
		Size theSize;
};

inline
Level operator -(const Level &l1, const Level &l2) {
	Level l = l1;
	l.theSize -= l2.theSize;
	l.theCount -= l2.theCount;
	return l;
}

// opens a file, fills it, closes
class Thread {
	public:
		Thread() { reset(); }
		~Thread() { if (theFD >= 0) close(); }

		int id() const { return theId; }
		bool done() const { return theDebt <= 0 && theFD < 0; }

		void id(int anId) { theId = anId; }
		void fileNo(int aFileNo) { theFileNo = aFileNo; }
		void size(Size aSize) { theSize = aSize; }

		inline void step();
	
	protected:
		void open();
		void write();
		void close();

		const char *fname();
		void reset();

	protected:
		Size theSize;  // file size
		Size theDebt;  // bytes left to write
		int theFileNo; // file "id"

		int theId;     // thread id
		int theFD;
};

// iterates among threads calling write() and fileNo()
class Loop {
	public:
		Loop();
		~Loop();
		
		void populate(int count);

		void step();

	protected:
		void Loop::refreshThread(Thread *th);
		Size Loop::calcFileSize(int objId);

	protected:
		SchArray<Thread*> theThreads;

		int theFileNo; // used to generate unique file numbers
};


static Array<const char *> TheTopDirs;
static int TheThreadCount = 0;
static Level TheLimit;
static Size TheFileSizeMean = Size::KB(4);

static const int L1 = 16;
static const int L2 = 64;

static const char TheIOBuf[8*1024];

static Level TheLevel;
static int TheIOCount = 0;
static int TheFileNoStep = 1;
static int TheFileNoShift = 0;
static int TheMaxFileNo = -1;

inline
void Thread::step() {
	Assert(theFileNo >= 0);
	if (theFD < 0)
		open();
	else
	if (theDebt > 0)
		write();
	else
		close();
}

void Thread::open() {
	const char *name = fname();
	if (theSize >= 0) {
		theFD = ::open(name, O_CREAT | O_WRONLY | O_TRUNC, 0644);
		if (theFD < 0) {
			cerr << "failed to open " << name << ": (" << errno << ") " << strerror(errno) << endl;
			exit(-1);
		}
	} else {
		if (::unlink(name) != 0) {
			cerr << "failed to unlink " << name << ": (" << errno << ") " << strerror(errno) << endl;
			// continue;
		}
			
	}
	theDebt = theSize;
}

void Thread::write() {
	size_t sz = ::write(theFD, TheIOBuf, min(theDebt, (Size)sizeof(TheIOBuf)));
	Check(sz > 0);
	theDebt -= sz;
	TheIOCount++;
	TheLevel.inc(0, sz);
}

void Thread::close() {
	Check(::close(theFD) == 0);
	theFD = -1;
	TheLevel.inc(1, 0);
}

const char *Thread::fname() {
	#define SWAP_DIR_SHIFT 24
	#define SWAP_FILE_MASK 0x00FFFFFF
	static char fullpath[4096];
	int dirn = (theFileNo >> SWAP_DIR_SHIFT) % TheTopDirs.count();
	int fn = theFileNo & SWAP_FILE_MASK;
	sprintf(fullpath, "%s/%02X/%02X/%08X",
		TheTopDirs[dirn],
		((fn / L2) / L2) % L1,
		(fn / L2) % L2,
		fn);
	return fullpath;
}

void Thread::reset() {
	theSize = theDebt = -1;
	theFD = theFileNo = -1;
}


/* Loop */

Loop::Loop() {
	theFileNo = TheFileNoShift;
}

Loop::~Loop() {
	for (int i = 0; i < theThreads.count(); ++i)
		delete theThreads[i];
}

void Loop::populate(int count) {

	for (int i = 0; i < count; ++i) {
		Thread *th = new Thread();
		th->id(i);
		theThreads.append(th);
		refreshThread(th);
	}
}

void Loop::step() {
	const int count = theThreads.count();
	for (int i = 0; i < count; ++i) {
		Thread *th = theThreads[i];
		th->step();
		if (th->done())
			refreshThread(th);
	}
}

void Loop::refreshThread(Thread *th) {
	static UnifDistr TheFileNoDistr(new RndGen(), 0, TheMaxFileNo);
	if (TheFileNoStep)
		theFileNo += TheFileNoStep;
	else
		theFileNo = (int) TheFileNoDistr();
	th->fileNo(theFileNo);
	th->size(calcFileSize(theFileNo));
}

Size Loop::calcFileSize(int objId) {
	// XXX: hard coded exponential distribution
	static ExpDistr TheSizeDistr(new RndGen(), TheFileSizeMean);

	if (objId == 1 || TheFileSizeMean < 0)
		return TheFileSizeMean;

	// seed random generator with objId
	TheSizeDistr.rndGen().seed((unsigned)objId);
	Size sz = (size_t) TheSizeDistr();
	// XXX: hard coded size limits:
	// most popular objects should not be too big (100*id + 10KB)
	// objects should not exceed 50MB
	sz %= min((Size)(100*objId + Size::KB(10)), Size::MB(50));
	// the object must be at least CBlock size
	return max(sz, (Size)32);

}

static
Size StrToSize(const char *s) {
	const int sz = strtol(s, (char**) &s, 10);
	if (!strcmp(s, "B"))
		return sz;
	if (!strcmp(s, "Kb"))
		return Size::KB(sz/8);
	if (!strcmp(s, "KB"))
		return Size::KB(sz);
	if (!strcmp(s, "Mb"))
		return Size::MB(sz/8);
	if (!strcmp(s, "MB"))
		return Size::MB(sz);
	if (!strcmp(s, "GB"))
		return Size::MB(1024*sz);
	cerr << "cannot parse size near " << s << endl;
	::abort();
	return 0;
}

static
void report(ostream &os, Time delay, const Level &level, int ioCount) {
	os << "filled: " << level.theCount << " files, " << level.theSize << " bytes, " << TheIOCount << " IOs in " << delay << endl;
	os << "rates: " << Rate(level.theCount, delay) << " files/sec, " << Rate(level.theSize/1024, delay) << " KB/sec, " << Rate(ioCount, delay) << " IO/sec" << endl;
}

int main(int argc, char *argv[]) {
	if (argc < 4 || !*argv[3]) {
		cerr << "usage: " << argv[0] << " <cache_dirs> <#threads> <#files|total_size> [mean_file_size]" << endl;
		return -1;
	}

	configureStream(cerr, 3);

	TheTopDirs.append(strtok(argv[1], ":"));
	while (const char *dir = strtok(NULL, ":"))
		TheTopDirs.append(dir);

	TheThreadCount = atoi(argv[2]);

	const char *ls = argv[3];
	if (isdigit(ls[strlen(ls)-1])) {
		TheLimit.theCount = atoi(ls);
		TheLimit.theSize = -1;
	} else {
		TheLimit.theSize = StrToSize(ls);
		TheLimit.theCount = -1;
	}

	if (argc > 4)
		TheFileSizeMean = StrToSize(argv[4]);

	if (const int realThCount = xatoi(getenv("FILLER_THREAD_COUNT"), -1) > -1) {
		TheFileNoStep = realThCount;
		TheFileNoShift = xatoi(getenv("FILLER_THREAD_ID"));
		cerr << "filler thread id: " << TheFileNoShift << ", skip factor: " << TheFileNoStep << endl;
		if (!TheFileNoStep)
			TheMaxFileNo = xatoi(getenv("FILLER_MAX_FILE_NO"), 10*1024);
	}
	

	cerr << "filling ";
	for (int i = 0; i < TheTopDirs.count(); ++i)
		cerr << (i ? ":" : "") << TheTopDirs[i];
	cerr << "; threads: " << TheThreadCount 
		<< "; limit: " << TheLimit.theCount << '/' << TheLimit.theSize 
		<< "; fsize: " << TheFileSizeMean << " bytes"
		<< endl;

	Loop TheLoop;
	TheLoop.populate(TheThreadCount);

	const Size TheReportSizeInterval = Size::KB(512);
    Size theNextReportSize = TheReportSizeInterval;

	const Time start = Time::Now();
	do {
		TheLoop.step();

		if (TheLevel.theSize >= theNextReportSize) {
			static Level lastLevel;
			static int lastIOCount = 0;
			report(cout, Time::Now() - start, TheLevel - lastLevel, TheIOCount - lastIOCount);
			while (TheLevel.theSize >= theNextReportSize)
				theNextReportSize += TheReportSizeInterval;
			lastLevel = TheLevel;
			lastIOCount = TheIOCount;
		}
	} while (TheLevel < TheLimit);
	
	report(cerr, Time::Now() - start, TheLevel, TheIOCount);
}
