// $Id: globals.cc,v 1.5.2.5 1998/10/16 02:20:46 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"
#include "globals.h"

#include <math.h>
#include <ctype.h>

World TheWorld;
Opt TheOpt;

int TheMaxFD = FD_SETSIZE; // XXX: this should be a configured value!
const char *ThePrgName = "PolyGraph";

/* World */

void World::syncZipfOmega() {
	// harmonic sum approximation: log(i) + gamma + 1/(2i)
	theZipfOmega = log(theCapacity) + TheEMConst + 1./(2*theCapacity);
}


/* Opt */

// set defaults
Opt::Opt(): theValPref(0), theValSuff(0) {

	theOriginAddr = Address("localhost", 8080);
	theProxyAddr = Address("127.0.0.1", 8080);

	theWorldId = 1;
	theWorldCapacity = (1 << 30);
	theSeqStart = -1;  // disabled
	theRndSeed = 1;    // default
	theUniqStart = -1; // disabled

	theObjSizeMean = Size::KB(13);

	theClientCount = 0;      // must be set
	theReqPerClient = 0;
	theClientThinkTime = -1; // disabled

	theServerThinkTime = -1; // disabled
	theServerCount = 0;
	isRepCachable = true;
}

Opt::~Opt() {
	delete[] theValPref;
	delete[] theValSuff;
}

bool Opt::parse(int argc, char *argv[]) {
	if (argc < 1+1) /* there is one required option */
		return false;

	int i;
	for (i = 1; i < argc; i += 2) {
		const char *name = argv[i];
		const char *val = argv[i+1];
		if (optEqual(name, "--host")) {
			theOriginAddr = splitVal(val) ?
				Address(theValPref, xatoi(theValSuff)) :
				Address(val, 80);
		} else
		if (optEqual(name, "--proxy") && splitVal(val))
			theProxyAddr = Address(theValPref, xatoi(theValSuff));
		else
		if (optEqual(name, "--world") && splitVal(val)) {
			theWorldId = xatoi(theValPref);
			theWorldCapacity = (strcasecmp(theValSuff, "max")) ?
				xatoi(theValSuff) : (1 << 30);
		} else
		if (optEqual(name, "--seq"))
			theSeqStart = xatoi(val, -1);
		else
		if (optEqual(name, "--rnd"))
			theRndSeed = xatoi(val, 1);
		else
		if (optEqual(name, "--unique"))
			theUniqStart = xatoi(val, -1);
		else
		if (optEqual(name, "--obj_size"))
			theObjSizeMean = Size::KB(xatoi(val));
		else
		if (optEqual(name, "--serv_think"))
			theServerThinkTime = xatoi(val, -1)/1000.;
		else
		if (optEqual(name, "--rep_cachable"))
			isRepCachable = xatoi(val, 1) > 0;
		else
		if (isdigit(*name))
			break;
		else {
			cerr << "unknown/broken option: '" << name << "'" << endl << endl;
			return false;
		}
	}

	if (!argv[i] || !isdigit(*argv[i]))
		return false;

	// process client specs
	theClientCount = xatoi(argv[i]);
	if (const char *h = strchr(argv[i], 'x')) { /* stream */
		theReqPerClient = xatoi(h+1);
	}
	if (const char *h = strchr(argv[i], 'r')) { /* poisson */
        const double req_per_second = xatoi(h+1);
		if (req_per_second > 0)
			theClientThinkTime = theClientCount/req_per_second; // [sec]
		else
			return false;
	}
	
	// server specs (listen ports)
	i++;
	theServerCount = argc - i;
	for (int p = 0; p < theServerCount; ++p)
		theServerPorts.append(xatoi(argv[i+p]));

	if (theClientCount > 0)
		return theReqPerClient > 0;
	else
		return theServerCount > 0;
}


void Opt::report(ostream &os) const {
	os << "configuration:" << endl
		<< "\tOriginAddr:\t " << theOriginAddr << endl
		<< "\tProxyAddr:\t " << theProxyAddr << endl
		<< "\tWorldId:\t " << theWorldId << endl
		<< "\tWorldCapacity:\t " << theWorldCapacity << " objects" << endl
		<< "\tSeqStart:\t " << theSeqStart << endl
		<< "\tRndSeed:\t " << theRndSeed << endl
		<< "\tUniqStart:\t " << theUniqStart << endl
		<< "\tObjSizeMean:\t " << theObjSizeMean << " bytes" << endl
		<< "\tClientCount:\t " << theClientCount << endl
		<< "\tReqPerClient:\t " << theReqPerClient << endl
		<< "\tClntThinkTime:\t " << theClientThinkTime << " seconds" <<endl
		<< "\tSvThinkTime:\t " << theServerThinkTime << " seconds" << endl
		<< "\tServerCount:\t " << theServerCount << endl
		<< "\tRepCachable:\t " << (isRepCachable ? "yes" : "no") << endl;

	os << "\tServerPorts:";
	for (int p = 0; p < theServerPorts.count(); ++p)
		os << "\t " << setw(4) << theServerPorts[p];
	os << endl;

	os << "workload:   " << (theClientThinkTime < 0 ? "stream" : "poisson") << endl;
	os << "obj_id_gen: " << (theRndSeed > 0 ? "random" : "sequential") << endl;
	os << "unique_url_gen: " << (theUniqStart < 0 ? "disabled" : "enabled") << endl;
}

bool Opt::optEqual(const char *name, const char *canonic) {
	if (!name)
		return false;
	
	// strip leading '-'
	while (*name == '-') name++;
	while (*canonic == '-') canonic++;

	if (!*name)
		return false;

	// matching prefix == equal! (sort of a kludge, be careful)
	return strncasecmp(name, canonic, strlen(name)) == 0;
}

bool Opt::splitVal(const char *value) {
	if (!value)
		return false;
	delete[] theValPref;
	delete[] theValSuff;
	theValPref = StrDup(value);
	if (char *h = strchr(theValPref, ':')) {
		*h = '\0';
		theValSuff = StrDup(h+1);
		return true;
	} else {
		cerr << "column (':') expected in '" << value << "'." << endl;
		theValSuff = StrDup("");
		return false;
	}	
}
